from __future__ import annotations

import gc
from typing import Union

import numpy as np
import torch
import vapoursynth as vs
from vapoursynth import core

from vsgan.constants import IS_VS_API_4, MAX_DTYPE_VALUES


def get_frame_plane(f: vs.VideoFrame, n: int, as_np: bool = False) -> Union[memoryview, np.array]:
    """
    Get a VideoFrame's Plane data as a MemoryView or a numpy array.
    Supports VS API 3 and 4.

    Parameters:
        f: VapourSynth VideoFrame from a clip.
        n: Plane number.
        as_np: Return as a Numpy Array.
    """
    if IS_VS_API_4:
        plane = f[n]
    else:
        plane = f.get_read_array(n)  # type: ignore
    if as_np:
        plane = np.asarray(plane)
    return plane


def frame_to_array(f: vs.VideoFrame) -> np.stack:
    """
    Convert a VapourSynth VideoFrame into a Numpy Array.

    Parameters:
        f: VapourSynth VideoFrame from a clip.
    """
    return np.stack([
        get_frame_plane(f, plane, as_np=True)
        for plane in range(f.format.num_planes)
    ])


def frame_to_tensor(f: vs.VideoFrame, order: tuple[int, ...] = (0, 1, 2), clamp_zero=True, bgr2rgb=False,
                    normalize=False, half: bool = False) -> torch.Tensor:
    """
    Convert a VapourSynth VideoFrame into a PyTorch Tensor.

    Parameters:
        f: VapourSynth VideoFrame from a clip.
        order: Change shape order to specified order. Default: CHW.
        clamp_zero: Clamp to 0,1 range.
        bgr2rgb: Flip Plane order from BGR to RGB. May be needed if loaded via OpenCV.
        normalize: Normalize (z-norm) from [0,1] range to [-1,1].
        half: Reduce tensor accuracy from fp32 to fp16. Reduces VRAM, may improve speed.
    """
    array = frame_to_array(f)

    if clamp_zero:
        max_val = MAX_DTYPE_VALUES.get(array.dtype, 1.0)
        array = array.astype(np.dtype("float32")) / max_val

    if order != (0, 1, 2) and len(order) == 3:
        array = np.transpose(array, order)

    array = torch.from_numpy(array).float()

    if half:
        array = array.half()

    if bgr2rgb:
        if array.shape[0] % 3 == 0:
            # RGB or MultixRGB (3xRGB, 5xRGB, etc. For video tensors.)
            array = array.flip(-3)
        elif array.shape[0] == 4:
            # RGBA
            array = array[[2, 1, 0, 3], :, :]

    if normalize:
        array = ((array - 0.5) * 2.0).clamp(-1, 1)

    return array


def tensor_to_frame(f: vs.VideoFrame, t: torch.Tensor) -> vs.VideoFrame:
    """
    Copies each channel from a Tensor into a VapourSynth VideoFrame.
    Supports any depth and format, and will return in the same format.

    It expects the tensor array to have the dimension count (C) first
    in the shape, e.g., CHW or CWH.

    Parameters:
        f: VapourSynth frame to store retrieved planes.
        t: PyTorch Tensor array to retrieve planes from.
    """
    # TODO: - Is the return needed? Looks like in-place modification
    #       - What if the frame is read-only?

    array = t.squeeze(0).detach().clamp(0, 1).cpu().numpy()

    d_type = np.asarray(f[0]).dtype
    array = MAX_DTYPE_VALUES.get(d_type, 1.0) * array
    array = array.astype(d_type)

    for plane in range(f.format.num_planes):
        d = np.asarray(f[plane])
        np.copyto(d, array[plane, :, :])

    return f


def tensor_to_clip(clip: vs.VideoNode, image: torch.Tensor) -> vs.VideoNode:
    """
    Convert a PyTorch Tensor into a VapourSynth VideoNode (clip).

    Expecting Torch shape to be in CHW order.

    Parameters:
        clip: Used to inherit expected return properties only.
        image: PyTorch Tensor.
    """
    clip = core.std.BlankClip(
        clip=clip,
        width=image.shape[-1],
        height=image.shape[-2]
    )
    return core.std.ModifyFrame(
        clip=clip,
        clips=clip,
        selector=lambda n, f: tensor_to_frame(f.copy(), image)
    )


def tile_tensor(t: torch.Tensor, overlap: int = 16) -> tuple[torch.Tensor, ...]:
    """
    Tile PyTorch Tensor into 4 quadrants with an overlap between tiles.
    Expects input PyTorch Tensor's shape to end in HW order.
    """
    b, c, h, w = t.shape

    top_left_lr = t[..., : h // 2 + overlap, : w // 2 + overlap]
    top_right_lr = t[..., : h // 2 + overlap, w // 2 - overlap:]
    bottom_left_lr = t[..., h // 2 - overlap:, : w // 2 + overlap]
    bottom_right_lr = t[..., h // 2 - overlap:, w // 2 - overlap:]

    return top_left_lr, top_right_lr, bottom_left_lr, bottom_right_lr


def recursive_tile_tensor(
    t: torch.Tensor,
    model: torch.nn.Module,
    overlap: int = 16,
    max_depth: int = None,
    current_depth: int = 1
) -> tuple[torch.Tensor, int]:
    """
    Recursively Tile PyTorch Tensor until the device has enough VRAM.
    It will try to tile as little as possible, and wont tile unless needed.
    Expects input PyTorch Tensor's shape to end in HW order.
    """
    if current_depth > 10:
        torch.cuda.empty_cache()
        gc.collect()
        raise RecursionError(f"Exceeded maximum tiling recursion of 10...")

    if max_depth is None or max_depth == current_depth:
        # attempt non-tiled super-resolution if no known depth, or at depth
        try:
            t_sr = model(t).data
            # del t  # TODO: Truly beneficial?
            return t_sr, current_depth
        except RuntimeError as e:
            if "allocate" in str(e) or "CUDA out of memory" in str(e):
                torch.cuda.empty_cache()
                gc.collect()  # TODO: Truly beneficial?
            else:
                raise

    # Not at known depth, and non-tiled super-resolution failed, try tiled

    # reduce overlap in half every recursion
    overlap //= current_depth

    tiles_lr = tile_tensor(t, overlap)
    # take depth from top_left result as the size would be same for all quadrants
    # by re-using the depth, we can know exactly how much tiling is needed immediately
    tiles_lr_top_left, depth = recursive_tile_tensor(tiles_lr[0], model, overlap, current_depth=current_depth + 1)
    tiles_lr_top_right, _ = recursive_tile_tensor(tiles_lr[1], model, overlap, depth, current_depth=current_depth + 1)
    tiles_lr_bottom_left, _ = recursive_tile_tensor(tiles_lr[2], model, overlap, depth, current_depth=current_depth + 1)
    tiles_lr_bottom_right, _ = recursive_tile_tensor(tiles_lr[3], model, overlap, depth, current_depth=current_depth + 1)

    output_img = join_tiles((tiles_lr_top_left, tiles_lr_top_right, tiles_lr_bottom_left, tiles_lr_bottom_right))
    return output_img, depth


def join_tiles(tiles_sr: tuple[torch.Tensor, ...], overlap: int = 16) -> torch.Tensor:
    """
    Join Tiled PyTorch Tensor quadrants into one large PyTorch Tensor.
    Expects input PyTorch Tensor's shapes to end in HW order.
    """
    b, c, h, w = tiles_sr[0].shape
    quadrants = len(tiles_sr)
    axis = quadrants // 2

    h = (h - (overlap * quadrants)) * axis
    w = (w - (overlap * quadrants)) * axis

    joined_tile = torch.empty((b, c, h, w), dtype=tiles_sr[0].dtype)
    joined_tile[..., : h // axis, : w // axis] = tiles_sr[0][..., : h // axis, : w // axis]
    joined_tile[..., : h // axis, -w // axis:] = tiles_sr[1][..., : h // axis, -w // axis:]
    joined_tile[..., -h // axis:, : w // axis] = tiles_sr[2][..., -h // axis:, : w // axis]
    joined_tile[..., -h // axis:, -w // axis:] = tiles_sr[3][..., -h // axis:, -w // axis:]

    return joined_tile



import functools

import numpy as np
import torch
import torch.nn as nn
# noinspection PyPep8Naming
import torch.nn.functional as F
from torch import Tensor

from vsgan.constants import STATE_T


class EGVSR(nn.Module):
    """
    EGVSR - Efficient & Generic Video Super-Resolution.
    By Yanpeng Cao, Chengcheng Wang, Changjun Song, Yongming Tang, and He Li.
    """

    def __init__(self, model: str, scale=4, in_nc=3, out_nc=3, nf=64, nb=16, degradation="BI") -> None:
        super(EGVSR, self).__init__()

        self.model = model
        self.state: STATE_T = torch.load(self.model)
        self.in_nc = in_nc
        self.out_nc = out_nc
        self.scale = scale
        self.num_filters = nf
        self.num_blocks = nb

        self.upsample_func = {
            "BI": functools.partial(F.interpolate, scale_factor=scale, mode="bilinear", align_corners=False),
            "BD": BicubicUpsample(scale_factor=scale)
        }.get(degradation)
        if self.upsample_func is None:
            raise NotImplementedError("Degradation [%s] is not recognized" % degradation)

        self.fnet = FNet(in_nc)
        self.srnet = SRNet(in_nc, out_nc, nf, nb, self.upsample_func)

        self.load_state_dict(self.state, strict=False)

    def forward(self, lr_curr, lr_prev, hr_prev):
        """
        Args:
            lr_curr: the current lr data in shape NCHW
            lr_prev: the previous lr data in shape NCHW
            hr_prev: the previous hr data in shape NC(4H)(4W)
        """

        # estimate lr flow (lr_curr -> lr_prev)
        lr_flow = self.fnet(lr_curr, lr_prev)

        # pad if size is not a multiple of 8
        pad_h = lr_curr.size(2) - lr_curr.size(2) // 8 * 8
        pad_w = lr_curr.size(3) - lr_curr.size(3) // 8 * 8
        lr_flow_pad = F.pad(lr_flow, (0, pad_w, 0, pad_h), 'reflect')

        # upsample lr flow
        hr_flow = self.scale * self.upsample_func(lr_flow_pad)

        # warp hr_prev
        hr_prev_warp = backward_warp(hr_prev, hr_flow)

        # compute hr_curr
        hr_curr = self.srnet(lr_curr, space_to_depth(hr_prev_warp, self.scale))

        return hr_curr

    def forward_sequence(self, lr_data: Tensor):
        """
        Args:
            lr_data: lr data in shape NTCHW
        """

        n, t, c, lr_h, lr_w = lr_data.size()
        hr_h, hr_w = lr_h * self.scale, lr_w * self.scale

        # calculate optical flows
        lr_prev = lr_data[:, :-1, ...].reshape(n * (t - 1), c, lr_h, lr_w)
        lr_curr = lr_data[:, 1:, ...].reshape(n * (t - 1), c, lr_h, lr_w)
        lr_flow = self.fnet(lr_curr, lr_prev)  # n*(t-1),2,h,w

        # upsample lr flows
        hr_flow = self.scale * self.upsample_func(lr_flow)
        hr_flow = hr_flow.view(n, (t - 1), 2, hr_h, hr_w)

        # compute the first hr data
        hr_data = []
        hr_prev = self.srnet(
            lr_data[:, 0, ...],
            torch.zeros(n, (self.scale ** 2) * c, lr_h, lr_w, dtype=torch.float32,
                        device=lr_data.device))
        hr_data.append(hr_prev)

        # compute the remaining hr data
        for i in range(1, t):
            # warp hr_prev
            hr_prev_warp = backward_warp(hr_prev, hr_flow[:, i - 1, ...])

            # compute hr_curr
            hr_curr = self.srnet(
                lr_data[:, i, ...],
                space_to_depth(hr_prev_warp, self.scale))

            # save and update
            hr_data.append(hr_curr)
            hr_prev = hr_curr

        hr_data = torch.stack(hr_data, dim=1)  # n,t,c,hr_h,hr_w

        # construct output dict
        """
        ret_dict = {
            'hr_data': hr_data,  # n,t,c,hr_h,hr_w
            'hr_flow': hr_flow,  # n,t,2,hr_h,hr_w
            'lr_prev': lr_prev,  # n(t-1),c,lr_h,lr_w
            'lr_curr': lr_curr,  # n(t-1),c,lr_h,lr_w
            'lr_flow': lr_flow,  # n(t-1),2,lr_h,lr_w
        }
        return ret_dict
        """
        return hr_data, hr_flow, lr_prev, lr_curr, lr_flow

    def infer_sequence(self, lr_data: torch.FloatTensor, device: torch.device) -> np.ndarray:
        """
        Args:
            lr_data: torch.FloatTensor in shape tchw
            device: torch.device

        Returns:
            hr_seq uint8 np.ndarray in shape tchw
        """

        # setup params
        tot_frm, c, h, w = lr_data.size()
        s = self.scale

        # forward
        hr_seq = []
        lr_prev = torch.zeros(1, c, h, w, dtype=torch.float32).to(device)
        hr_prev = torch.zeros(
            1, c, s * h, s * w, dtype=torch.float32).to(device)

        for i in range(tot_frm):
            with torch.no_grad():
                self.eval()

                lr_curr = lr_data[i: i + 1, ...].to(device)
                hr_curr = self.forward(lr_curr, lr_prev, hr_prev)
                lr_prev, hr_prev = lr_curr, hr_curr

                hr_frm = hr_curr.squeeze(0).cpu().numpy()  # chw|rgb|uint8

            hr_seq.append(float32_to_uint8(hr_frm))

        return np.stack(hr_seq).transpose(0, 2, 3, 1)  # thwc


class FNet(nn.Module):
    """Optical flow estimation network."""

    def __init__(self, in_nc: int):
        super(FNet, self).__init__()

        self.encoder1 = nn.Sequential(
            nn.Conv2d(2 * in_nc, 32, 3, 1, 1, bias=True),
            nn.LeakyReLU(0.2, inplace=True),
            nn.Conv2d(32, 32, 3, 1, 1, bias=True),
            nn.LeakyReLU(0.2, inplace=True),
            nn.MaxPool2d(2, 2))

        self.encoder2 = nn.Sequential(
            nn.Conv2d(32, 64, 3, 1, 1, bias=True),
            nn.LeakyReLU(0.2, inplace=True),
            nn.Conv2d(64, 64, 3, 1, 1, bias=True),
            nn.LeakyReLU(0.2, inplace=True),
            nn.MaxPool2d(2, 2))

        self.encoder3 = nn.Sequential(
            nn.Conv2d(64, 128, 3, 1, 1, bias=True),
            nn.LeakyReLU(0.2, inplace=True),
            nn.Conv2d(128, 128, 3, 1, 1, bias=True),
            nn.LeakyReLU(0.2, inplace=True),
            nn.MaxPool2d(2, 2))

        self.decoder1 = nn.Sequential(
            nn.Conv2d(128, 256, 3, 1, 1, bias=True),
            nn.LeakyReLU(0.2, inplace=True),
            nn.Conv2d(256, 256, 3, 1, 1, bias=True),
            nn.LeakyReLU(0.2, inplace=True))

        self.decoder2 = nn.Sequential(
            nn.Conv2d(256, 128, 3, 1, 1, bias=True),
            nn.LeakyReLU(0.2, inplace=True),
            nn.Conv2d(128, 128, 3, 1, 1, bias=True),
            nn.LeakyReLU(0.2, inplace=True))

        self.decoder3 = nn.Sequential(
            nn.Conv2d(128, 64, 3, 1, 1, bias=True),
            nn.LeakyReLU(0.2, inplace=True),
            nn.Conv2d(64, 64, 3, 1, 1, bias=True),
            nn.LeakyReLU(0.2, inplace=True))

        self.flow = nn.Sequential(
            nn.Conv2d(64, 32, 3, 1, 1, bias=True),
            nn.LeakyReLU(0.2, inplace=True),
            nn.Conv2d(32, 2, 3, 1, 1, bias=True))

    def forward(self, x1, x2):
        """Compute optical flow from x1 to x2."""
        out = self.encoder1(torch.cat([x1, x2], dim=1))
        out = self.encoder2(out)
        out = self.encoder3(out)
        out = F.interpolate(self.decoder1(out), scale_factor=2, mode='bilinear', align_corners=False)
        out = F.interpolate(self.decoder2(out), scale_factor=2, mode='bilinear', align_corners=False)
        out = F.interpolate(self.decoder3(out), scale_factor=2, mode='bilinear', align_corners=False)
        out = torch.tanh(self.flow(out)) * 24  # 24 is the max velocity
        return out


class SRNet(nn.Module):
    """Reconstruction and Upsampling network."""

    def __init__(self, in_nc: int = 3, out_nc: int = 3, nf: int = 64, nb: int = 16, upsample_func=None, scale: int = 4):
        super(SRNet, self).__init__()

        # input conv.
        self.conv_in = nn.Sequential(
            nn.Conv2d((scale ** 2 + 1) * in_nc, nf, 3, 1, 1, bias=True),
            nn.ReLU(inplace=True))

        # residual blocks
        self.resblocks = nn.Sequential(*[ResidualBlock(nf) for _ in range(nb)])

        # upsampling
        self.conv_up = nn.Sequential(
            nn.ConvTranspose2d(nf, nf, 3, 2, 1, output_padding=1, bias=True),
            nn.ReLU(inplace=True),
            nn.ConvTranspose2d(nf, nf, 3, 2, 1, output_padding=1, bias=True),
            nn.ReLU(inplace=True))

        self.conv_up_cheap = nn.Sequential(
            nn.PixelShuffle(4),
            nn.ReLU(inplace=True))

        # output conv.
        self.conv_out = nn.Conv2d(4, out_nc, 3, 1, 1, bias=True)

        # upsampling function
        self.upsample_func = upsample_func

    def forward(self, lr_curr, hr_prev_tran):
        """ lr_curr: the current lr data in shape nchw
            hr_prev_tran: the previous transformed hr_data in shape n(4*4*c)hw
        """

        out = self.conv_in(torch.cat([lr_curr, hr_prev_tran], dim=1))
        out = self.resblocks(out)
        out = self.conv_up_cheap(out)
        out = self.conv_out(out)
        # out += self.upsample_func(lr_curr)

        return out


class ResidualBlock(nn.Module):
    """Residual block without batch normalization."""

    def __init__(self, nf: int = 64):
        super(ResidualBlock, self).__init__()
        self.conv = nn.Sequential(
            nn.Conv2d(nf, nf, 3, 1, 1, bias=True),
            nn.ReLU(inplace=True),
            nn.Conv2d(nf, nf, 3, 1, 1, bias=True)
        )

    def forward(self, x):
        return self.conv(x) + x


class BicubicUpsample(nn.Module):
    """
    A bicubic upsampling class with similar behavior to that in TecoGAN-Tensorflow.

    Note that it's different from torch.nn.functional.interpolate and matlab's
    imresize in terms of bicubic kernel and sampling scheme.

    Theoretically it can support any scale_factor >= 1, but currently only
    scale_factor = 4 is tested.

    References:
        http://verona.fi-p.unam.mx/boris/practicas/CubConvInterp.pdf
        https://stackoverflow.com/q/26823140/13183782
    """

    def __init__(self, scale_factor, a=-0.75):
        super(BicubicUpsample, self).__init__()

        # calculate weights
        cubic = torch.FloatTensor([
            [0, a, -2 * a, a],
            [1, 0, -(a + 3), a + 2],
            [0, -a, (2 * a + 3), -(a + 2)],
            [0, 0, a, -a]
        ])  # accord to Eq.(6) in the reference paper

        kernels = [
            torch.matmul(cubic, torch.FloatTensor([1, s, s ** 2, s ** 3]))
            for s in [1.0 * d / scale_factor for d in range(scale_factor)]
        ]  # s = x - floor(x)

        # register parameters
        self.scale_factor = scale_factor
        self.register_buffer('kernels', torch.stack(kernels))

    def forward(self, input):
        n, c, h, w = input.size()
        s = self.scale_factor

        # pad input (left, right, top, bottom)
        input = F.pad(input, (1, 2, 1, 2), mode='replicate')

        # calculate output (height)
        kernel_h = self.kernels.repeat(c, 1).view(-1, 1, s, 1)
        output = F.conv2d(input, kernel_h, stride=1, padding=0, groups=c)
        output = output.reshape(
            n, c, s, -1, w + 3).permute(0, 1, 3, 2, 4).reshape(n, c, -1, w + 3)

        # calculate output (width)
        kernel_w = self.kernels.repeat(c, 1).view(-1, 1, 1, s)
        output = F.conv2d(output, kernel_w, stride=1, padding=0, groups=c)
        output = output.reshape(
            n, c, s, h * s, -1).permute(0, 1, 3, 4, 2).reshape(n, c, h * s, -1)

        return output


def backward_warp(x: Tensor, flow: Tensor, mode: str = "bilinear", padding_mode: str = "border") -> Tensor:
    """
    Backward warp `x` according to `flow`.
    Both x and flow are pytorch tensor in shape `nchw` and `n2hw`.
    Reference:
        https://github.com/sniklaus/pytorch-spynet/blob/master/run.py#L41
    """

    n, c, h, w = x.size()

    # create mesh grid
    iu = torch.linspace(-1.0, 1.0, w).view(1, 1, 1, w).expand(n, -1, h, -1)
    iv = torch.linspace(-1.0, 1.0, h).view(1, 1, h, 1).expand(n, -1, -1, w)
    grid = torch.cat([iu, iv], 1).to(flow.device)

    # normalize flow to [-1, 1]
    flow = torch.cat([
        flow[:, 0:1, ...] / ((w - 1.0) / 2.0),
        flow[:, 1:2, ...] / ((h - 1.0) / 2.0)], dim=1)

    # add flow to grid and reshape to nhw2
    grid = (grid + flow).permute(0, 2, 3, 1)

    # bilinear sampling
    output = F.grid_sample(x, grid, mode=mode, padding_mode=padding_mode, align_corners=True)

    return output


def space_to_depth(x: Tensor, scale: int = 4) -> Tensor:
    """Equivalent to tf.space_to_depth()."""
    n, c, in_h, in_w = x.size()
    out_h, out_w = in_h // scale, in_w // scale

    x_reshaped = x.reshape(n, c, out_h, scale, out_w, scale)
    x_reshaped = x_reshaped.permute(0, 3, 5, 1, 2, 4)
    output = x_reshaped.reshape(n, scale * scale * c, out_h, out_w)

    return output


def float32_to_uint8(array: np.ndarray) -> np.ndarray:
    """
    Convert np.float32 array to np.uint8.

    Args:
        array: np.float32, (NT)CHW, [0, 1]

    Returns:
        np.uint8, (NT)CHW, [0, 255]
    """
    return np.uint8(np.clip(np.round(array * 255), 0, 255))



from __future__ import annotations

from typing import OrderedDict

import numpy as np
import vapoursynth as vs
from torch import Tensor

VS_API = vs.__api_version__.api_major
IS_VS_API_4 = VS_API == 4

MAX_DTYPE_VALUES = {
    np.dtype("int8"): 127,
    np.dtype("uint8"): 255,
    np.dtype("int16"): 32767,
    np.dtype("uint16"): 65535,
    np.dtype("int32"): 2147483647,
    np.dtype("uint32"): 4294967295,
    np.dtype("int64"): 9223372036854775807,
    np.dtype("uint64"): 18446744073709551615,
    np.dtype("float32"): 1.0,
    np.dtype("float64"): 1.0,
}
STATE_T = OrderedDict[str, Tensor]



from __future__ import annotations

from abc import abstractmethod
from typing import Union, Optional

import torch
import vapoursynth as vs


class BaseNetwork:
    def __init__(self, clip: vs.VideoNode, device: Union[str, int] = "cuda"):
        """
        Create a PyTorch Device instance to use VSGAN with.
        It validates the supplied pytorch device identifier, and makes
        sure CUDA environment is available and ready.

        Parameters:
            clip: VapourSynth Video Node (aka clip). It must be RGB
                color-space. RGB27 and RGB30 may not work.
            device: PyTorch device identifier to use for the model. E.g.,
                "cuda", "cpu", 0, 1, and so on.
        """
        if not isinstance(clip, vs.VideoNode):
            raise ValueError(f"This is not a clip, {clip!r}")

        if clip.format.color_family.name != "RGB":
            raise ValueError("Only RGB clips are supported. RGB24 or RGBS recommended.")

        device = device.strip().lower() if isinstance(device, str) else device
        if device == "":
            raise ValueError("Device must be provided.")
        if device == "cpu":
            raise UserWarning(
                "VSGAN blocked an attempt to use your CPU as the torch device. "
                "Using your CPU will run it at very high utilisation and may lower its lifespan. "
                "If you are sure you would like to use your CPU, then use `cpu!`."
            )
        if device == "cpu!":
            device = "cpu"
        if device != "cpu" and not torch.cuda.is_available():
            raise EnvironmentError("Either NVIDIA CUDA or the device (%s) isn't available." % device)

        self.clip: vs.VideoNode = clip
        self.device: torch.device = torch.device(device)
        self.model: Optional[torch.nn.Module] = None
        self.half: bool = False

    @abstractmethod
    def load(self, model: str, half: bool = False) -> BaseNetwork:
        """
        Load a PyTorch model file and send to the PyTorch device.
        The model can be changed at any point.

        Parameters:
            model: Path to a supported PyTorch Model file.
            half: Reduce tensor accuracy from fp32 to fp16. Reduces VRAM, may improve speed.
        """

    @abstractmethod
    def apply(self, overlap: int = 16) -> BaseNetwork:
        """
        Apply the model on each frame of the clip.

        Overlap should generally be a multiple of 16. The larger the input resolution,
        the larger overlap may need to be set. Avoid using a value excessively large.

        Parameters:
            overlap: Amount to overlap each tile as to hide artefact seams.
        """



from __future__ import annotations

import functools
from typing import Union

import torch
import vapoursynth as vs
from vapoursynth import core

from vsgan import archs
from vsgan.networks.basenetwork import BaseNetwork
from vsgan.utilities import frame_to_tensor, tensor_to_clip


class EGVSR(BaseNetwork):
    """
    EGVSR - Efficient & Generic Video Super-Resolution.
    By Yanpeng Cao, Chengcheng Wang, Changjun Song, Yongming Tang, and He Li.
    """

    def __init__(self, clip: vs.VideoNode, device: Union[str, int] = "cuda"):
        super().__init__(clip, device)
        self.tensor_cache: dict = {}

    def load(
        self,
        model: str,
        scale: int = 4,
        in_nc: int = 3,
        out_nc: int = 3,
        nf: int = 64,
        nb: int = 16,
        degradation: str = "BI",
        half: bool = False
    ) -> EGVSR:
        """
        Load an EGVSR model file and send to the PyTorch device.
        The model can be changed at any point.

        Supported Model Files:
        - Must be a Generator model.
        - EGVSR only.

        Parameters:
            model: Path to a supported PyTorch Model file.
            scale: Model Scale, the resulting scale relative to the input.
            in_nc: Input number of channels.
            out_nc: Output number of channels.
            nf: Number of filters.
            nb: Number of blocks.
            degradation: Upsample Function.
            half: Reduce tensor accuracy from fp32 to fp16. Reduces VRAM, may improve speed.
        """
        model = archs.EGVSR(model, scale, in_nc, out_nc, nf, nb, degradation)
        model.eval()
        self.model = model.to(self.device)
        if half:
            self.model.half()
        self.half = half
        return self

    @torch.inference_mode()
    def apply(self, interval: int = 5) -> EGVSR:
        """
        Apply the model on each frame of the clip.

        Parameters:
            interval: Amount of frames ahead to inference. Must be greater than 0.
        """
        if not self.model:
            raise ValueError("A model must be loaded before running.")

        def _apply(n: int, clip: vs.VideoNode, model: torch.nn.Module, half: bool, interval_: int) -> vs.VideoNode:
            if str(n) not in self.tensor_cache:
                self.tensor_cache.clear()

                lr_images = [frame_to_tensor(clip.get_frame(n))]
                for i in range(1, interval_):
                    if (n + i) >= clip.num_frames:
                        break
                    lr_images.append(frame_to_tensor(clip.get_frame(n + i)))
                lr_images = torch.stack(lr_images)
                lr_images = lr_images.unsqueeze(0)
                if half:
                    lr_images = lr_images.half()

                output, _, _, _, _ = model.forward_sequence(lr_images.to(self.device))
                output = output.squeeze(0)

                for i in range(output.shape[0]):  # interval
                    self.tensor_cache[str(n + i)] = output[i, :, :, :]

                del lr_images
                torch.cuda.empty_cache()

            return tensor_to_clip(clip, self.tensor_cache[str(n)])

        self.clip = core.std.FrameEval(
            core.std.BlankClip(
                clip=self.clip,
                width=self.clip.width * self.model.scale,
                height=self.clip.height * self.model.scale
            ),
            functools.partial(
                _apply,
                clip=self.clip,
                model=self.model,
                half=self.half,
                interval_=interval
            )
        )

        return self


__ALL__ = (EGVSR,)



from typing import Union

from vsgan.archs.EGVSR import EGVSR
from vsgan.archs.ESRGAN import ESRGAN, RealESRGANv2

MODEL_T = Union[ESRGAN, RealESRGANv2, EGVSR]

__ALL__ = (MODEL_T, ESRGAN, RealESRGANv2, EGVSR)



from __future__ import annotations

import functools

import torch
import vapoursynth as vs
from vapoursynth import core

from vsgan import archs
from vsgan.networks.basenetwork import BaseNetwork
from vsgan.utilities import frame_to_tensor, tensor_to_clip, recursive_tile_tensor


class ESRGAN(BaseNetwork):
    """
    ESRGAN - Enhanced Super-Resolution Generative Adversarial Networks.
    By Xintao Wang, Ke Yu, Shixiang Wu, Jinjin Gu, Yihao Liu, Chao Dong, Yu Qiao,
    and Chen Change Loy.
    """

    def load(self, model: str, half: bool = False) -> ESRGAN:
        """
        Load an ESRGAN model file and send to the PyTorch device.
        The model can be changed at any point.

        Supported Model Files:
        - Must be a Generator model.
        - ESRGAN (old and new)
        - ESRGAN+
        - Real-ESRGAN (v1 and v2)
        - A-ESRGAN

        Parameters:
            model: Path to a supported PyTorch Model file.
            half: Reduce tensor accuracy from fp32 to fp16. Reduces VRAM, may improve speed.
        """
        state = torch.load(model)
        if "params" in state and "body.0.weight" in state["params"]:
            arch = archs.RealESRGANv2
        else:
            arch = archs.ESRGAN
        model = arch(state)
        model.eval()
        self.model = model.to(self.device)
        if half:
            self.model.half()
        self.half = half
        return self

    @torch.inference_mode()
    def apply(self, overlap: int = 16) -> ESRGAN:
        """
        Apply the model on each frame of the clip.

        Overlap should generally be a multiple of 16. The larger the input resolution,
        the larger overlap may need to be set. Avoid using a value excessively large.

        Parameters:
            overlap: Amount to overlap each tile as to hide artefact seams.
        """
        if not self.model:
            raise ValueError("A model must be loaded before running.")

        def _apply(n: int, clip: vs.VideoNode, model: torch.nn.Module, half: bool, overlap_: int) -> vs.VideoNode:
            lr_img = frame_to_tensor(clip.get_frame(n), half=half)
            lr_img.unsqueeze_(0)
            output_img, depth = recursive_tile_tensor(lr_img.to(self.device), model, overlap_)
            return tensor_to_clip(clip, output_img)

        self.clip = core.std.FrameEval(
            core.std.BlankClip(
                clip=self.clip,
                width=self.clip.width * self.model.scale,
                height=self.clip.height * self.model.scale
            ),
            functools.partial(
                _apply,
                clip=self.clip,
                model=self.model,
                half=self.half,
                overlap_=overlap
            )
        )

        return self


__ALL__ = (ESRGAN,)



import math
import re
from collections import OrderedDict
from typing import Optional

import torch
import torch.nn as nn
# noinspection PyPep8Naming
import torch.nn.functional as F
from torch import pixel_unshuffle

from vsgan.archs import blocks as block
from vsgan.constants import STATE_T


class ESRGAN(nn.Module):
    """
    ESRGAN - Enhanced Super-Resolution Generative Adversarial Networks.
    By Xintao Wang, Ke Yu, Shixiang Wu, Jinjin Gu, Yihao Liu, Chao Dong, Yu Qiao,
    and Chen Change Loy.
    """

    def __init__(self, state: STATE_T, norm=None, act: str = "leakyrelu", upsampler: str = "upconv",
                 mode: str = "CNA") -> None:
        """
        This is old-arch Residual in Residual Dense Block Network and is not
        the newest revision that's available at github.com/xinntao/ESRGAN.
        This is on purpose, the newest Network has severely limited the
        potential use of the Network with no benefits.

        This network supports model files from both new and old-arch.

        Args:
            state: PyTorch Model State dictionary.
            norm: Normalization layer
            act: Activation layer
            upsampler: Upsample layer. upconv, pixel_shuffle
            mode: Convolution mode
        """
        super(ESRGAN, self).__init__()

        self.state = state
        self.norm = norm
        self.act = act
        self.upsampler = upsampler
        self.mode = mode

        self.state_map = {
            # wanted, possible key names
            # currently supports old, new, and newer RRDBNet arch models
            "model.0.weight": ("conv_first.weight",),
            "model.0.bias": ("conv_first.bias",),
            "model.1.sub./NB/.weight": ("trunk_conv.weight", "conv_body.weight"),
            "model.1.sub./NB/.bias": ("trunk_conv.bias", "conv_body.bias"),
            "model.3.weight": ("upconv1.weight", "conv_up1.weight"),
            "model.3.bias": ("upconv1.bias", "conv_up1.bias"),
            "model.6.weight": ("upconv2.weight", "conv_up2.weight"),
            "model.6.bias": ("upconv2.bias", "conv_up2.bias"),
            "model.8.weight": ("HRconv.weight", "conv_hr.weight"),
            "model.8.bias": ("HRconv.bias", "conv_hr.bias"),
            "model.10.weight": ("conv_last.weight",),
            "model.10.bias": ("conv_last.bias",),
            r"model.1.sub.\1.RDB\2.conv\3.0.\4": (
                r"RRDB_trunk\.(\d+)\.RDB(\d)\.conv(\d+)\.(weight|bias)",
                r"body\.(\d+)\.rdb(\d)\.conv(\d+)\.(weight|bias)"
            )
        }

        if "params_ema" in self.state:
            self.state = self.state["params_ema"]

        self.num_blocks = self.get_num_blocks()
        self.plus = any("conv1x1" in k for k in self.state.keys())

        self.state: STATE_T = self.new_to_old_arch(self.state)
        self.in_nc = self.state["model.0.weight"].shape[1]
        self.out_nc = self.get_out_nc() or self.in_nc  # assume same as in nc if not found
        self.scale = self.get_scale()
        self.num_filters = self.state["model.0.weight"].shape[0]

        if self.in_nc in (self.out_nc * 4, self.out_nc * 16) and \
                self.out_nc in (self.in_nc / 4, self.in_nc / 16):
            self.shuffle_factor = int(math.sqrt(self.in_nc / self.out_nc))
        else:
            self.shuffle_factor = None

        upsample_block = {
            "upconv": block.upconv_block,
            "pixel_shuffle": block.pixelshuffle_block
        }.get(self.upsampler)
        if upsample_block is None:
            raise NotImplementedError("Upsample mode [%s] is not found" % self.upsampler)

        if self.scale == 3:
            upsample_blocks = upsample_block(
                in_nc=self.num_filters,
                out_nc=self.num_filters,
                upscale_factor=3,
                act_type=self.act
            )
        else:
            upsample_blocks = [upsample_block(
                in_nc=self.num_filters,
                out_nc=self.num_filters,
                act_type=self.act
            ) for _ in range(int(math.log(self.scale, 2)))]

        self.model = block.sequential(
            # fea conv
            block.conv_block(
                in_nc=self.in_nc,
                out_nc=self.num_filters,
                kernel_size=3,
                norm_type=None,
                act_type=None
            ),
            block.ShortcutBlock(block.sequential(
                # rrdb blocks
                *[RRDB(
                    nc=self.num_filters,
                    kernel_size=3,
                    gc=32,
                    stride=1,
                    bias=True,
                    pad_type="zero",
                    norm_type=self.norm,
                    act_type=self.act,
                    mode="CNA",
                    plus=self.plus
                ) for _ in range(self.num_blocks)],
                # lr conv
                block.conv_block(
                    in_nc=self.num_filters,
                    out_nc=self.num_filters,
                    kernel_size=3,
                    norm_type=self.norm,
                    act_type=None,
                    mode=self.mode
                )
            )),
            *upsample_blocks,
            # hr_conv0
            block.conv_block(
                in_nc=self.num_filters,
                out_nc=self.num_filters,
                kernel_size=3,
                norm_type=None,
                act_type=self.act
            ),
            # hr_conv1
            block.conv_block(
                in_nc=self.num_filters,
                out_nc=self.out_nc,
                kernel_size=3,
                norm_type=None,
                act_type=None
            )
        )

        # vapoursynth calls expect the real scale even if shuffled
        if self.shuffle_factor:
            self.scale = self.scale // self.shuffle_factor

        self.load_state_dict(self.state, strict=False)

    def new_to_old_arch(self, state: STATE_T) -> STATE_T:
        """Convert a new-arch model state dictionary to an old-arch dictionary."""
        if "params_ema" in state:
            state = state["params_ema"]

        if "conv_first.weight" not in state:
            # model is already old arch, this is a loose check, but should be sufficient
            return state

        # add nb to state keys
        for kind in ("weight", "bias"):
            self.state_map[f"model.1.sub.{self.num_blocks}.{kind}"] = self.state_map[f"model.1.sub./NB/.{kind}"]
            del self.state_map[f"model.1.sub./NB/.{kind}"]

        old_state = OrderedDict()
        for old_key, new_keys in self.state_map.items():
            for new_key in new_keys:
                if r"\1" in old_key:
                    for k, v in state.items():
                        sub = re.sub(new_key, old_key, k)
                        if sub != k:
                            old_state[sub] = v
                else:
                    if new_key in state:
                        old_state[old_key] = state[new_key]

        return old_state

    def get_out_nc(self) -> Optional[int]:
        max_part = 0
        out_nc = None
        for part in list(self.state):
            parts = part.split(".")[1:]
            if len(parts) == 2:
                part_num = int(parts[0])
                if part_num > max_part:
                    max_part = part_num
                    out_nc = self.state[part].shape[0]
        return out_nc

    def get_scale(self, min_part: int = 6) -> int:
        n = 0
        for part in list(self.state):
            parts = part.split(".")[1:]
            if len(parts) == 2:
                part_num = int(parts[0])
                if part_num > min_part and parts[1] == "weight":
                    n += 1
        return 2 ** n

    def get_num_blocks(self) -> int:
        nbs = []
        state_keys = self.state_map[r"model.1.sub.\1.RDB\2.conv\3.0.\4"] + (
            r"model\.\d+\.sub\.(\d+)\.RDB(\d+)\.conv(\d+)\.0\.(weight|bias)",
        )
        for state_key in state_keys:
            for k in self.state:
                m = re.search(state_key, k)
                if m:
                    nbs.append(int(m.group(1)))
            if nbs:
                break
        return max(*nbs) + 1

    def forward(self, x):
        if self.shuffle_factor:
            x = pixel_unshuffle(x, downscale_factor=self.shuffle_factor)
        return self.model(x)


class RealESRGANv2(nn.Module):
    def __init__(self, state: STATE_T, act_type: str = "prelu") -> None:
        """
        Real-ESRGAN - Training Real-World Blind Super-Resolution with Pure Synthetic Data.
        By Xintao Wang, Liangbin Xie, Chao Dong, and Ying Shan.

        This is specifically the alternate version used mainly for video inference.
        However, this is still not a video super-resolution network as it still only takes in
        one frame (or image) at a time. See:
        https://github.com/xinntao/Real-ESRGAN/blob/master/docs/model_zoo.md#for-anime-videos

        This is a compact VGG-style network structure for super-resolution, which performs
        upsampling in the last layer. No convolution is conducted on the HR feature space.

        Args:
            state: PyTorch Model State dictionary.
            act_type: Activation type: 'relu', 'prelu', 'leakyrelu'.
        """
        super(RealESRGANv2, self).__init__()

        self.state = state
        self.act_type = act_type

        if "params" in self.state:
            # TODO: Should *all* Real-ESRGAN v2 models begin with "params" key?
            #       Reject models that don't?
            self.state = self.state["params"]

        self.state_keys = list(self.state.keys())

        self.num_in_ch = self.get_in_nc()
        self.num_out_ch = self.num_in_ch  # TODO: Find a way. Assuming same as in_nc...
        self.num_conv = self.get_num_conv()
        self.num_feat = self.get_num_feats()
        self.pixel_shuffle_shape = self.get_pixel_shuffle_shape()
        self.scale = self.get_scale()

        # activation
        activation = {
            "relu": nn.ReLU(inplace=True),
            "prelu": nn.PReLU(num_parameters=self.num_feat),
            "leakyrelu": nn.LeakyReLU(negative_slope=0.1, inplace=True)
        }.get(self.act_type, None)
        if activation is None:
            raise ValueError("Activation type [%s] was unrecognized" % self.act_type)

        # body structure
        self.body = nn.ModuleList()
        for n in range(self.num_conv + 1):
            self.body.append(nn.Conv2d(self.num_in_ch if n == 0 else self.num_feat, self.num_feat, 3, 1, 1))
            self.body.append({
                "relu": nn.ReLU(inplace=True),
                "prelu": nn.PReLU(num_parameters=self.num_feat),
                "leakyrelu": nn.LeakyReLU(negative_slope=0.1, inplace=True)
            }[self.act_type])

        # last conv
        self.body.append(nn.Conv2d(self.num_feat, self.pixel_shuffle_shape, 3, 1, 1))

        # upsampler
        self.upsampler = nn.PixelShuffle(self.scale)

        self.load_state_dict(self.state, strict=False)

    def get_in_nc(self) -> int:
        return self.state[self.state_keys[0]].shape[1]

    def get_num_conv(self) -> int:
        return (int(self.state_keys[-1].split(".")[1]) - 2) // 2

    def get_num_feats(self) -> int:
        return self.state[self.state_keys[0]].shape[0]

    def get_pixel_shuffle_shape(self) -> int:
        return self.state[self.state_keys[-1]].shape[0]

    def get_scale(self) -> int:
        # Assume out_nc is the same as in_nc
        # I cant think of a better way to do that
        self.num_out_ch = self.num_in_ch
        scale = math.sqrt(self.pixel_shuffle_shape / self.num_out_ch)
        if scale - int(scale) > 0:
            print("out_nc is probably different than in_nc, scale calculation might be wrong")
        scale = int(scale)
        return scale

    def forward(self, x):
        out = x
        for i in range(0, len(self.body)):
            out = self.body[i](out)

        out = self.upsampler(out)
        # add the nearest upsampled image, so that the network learns the residual
        base = F.interpolate(x, scale_factor=self.scale, mode="nearest")
        out += base
        return out


class ResidualDenseBlock5C(nn.Module):
    """
    5 Convolution Residual Dense Block.
    Residual Dense Network for Image Super-Resolution, CVPR 18.
    gc: growth channel, i.e. intermediate channels
    """

    def __init__(self, nc, kernel_size=3, gc=32, stride=1, bias=True, pad_type="zero", norm_type=None,
                 act_type="leakyrelu", mode="CNA", plus=False):
        super(ResidualDenseBlock5C, self).__init__()
        last_act = None if mode == "CNA" else act_type

        self.conv1x1 = conv1x1(nc, gc) if plus else None
        self.conv1 = block.conv_block(nc, gc, kernel_size, stride, bias=bias, pad_type=pad_type,
                                      norm_type=norm_type, act_type=act_type, mode=mode)
        self.conv2 = block.conv_block(nc + gc, gc, kernel_size, stride, bias=bias, pad_type=pad_type,
                                      norm_type=norm_type, act_type=act_type, mode=mode)
        self.conv3 = block.conv_block(nc + 2 * gc, gc, kernel_size, stride, bias=bias, pad_type=pad_type,
                                      norm_type=norm_type, act_type=act_type, mode=mode)
        self.conv4 = block.conv_block(nc + 3 * gc, gc, kernel_size, stride, bias=bias, pad_type=pad_type,
                                      norm_type=norm_type, act_type=act_type, mode=mode)
        self.conv5 = block.conv_block(nc + 4 * gc, nc, 3, stride, bias=bias, pad_type=pad_type,
                                      norm_type=norm_type, act_type=last_act, mode=mode)

    def forward(self, x):
        x1 = self.conv1(x)
        x2 = self.conv2(torch.cat((x, x1), 1))
        if self.conv1x1:
            x2 = x2 + self.conv1x1(x)
        x3 = self.conv3(torch.cat((x, x1, x2), 1))
        x4 = self.conv4(torch.cat((x, x1, x2, x3), 1))
        if self.conv1x1:
            x4 = x4 + x2
        x5 = self.conv5(torch.cat((x, x1, x2, x3, x4), 1))
        return x5.mul(0.2) + x


class RRDB(nn.Module):
    """Residual in Residual Dense Block."""

    def __init__(self, nc, kernel_size=3, gc=32, stride=1, bias=True, pad_type="zero", norm_type=None,
                 act_type="leakyrelu", mode="CNA", plus=False):
        super(RRDB, self).__init__()
        self.RDB1 = ResidualDenseBlock5C(nc, kernel_size, gc, stride, bias, pad_type, norm_type, act_type, mode, plus)
        self.RDB2 = ResidualDenseBlock5C(nc, kernel_size, gc, stride, bias, pad_type, norm_type, act_type, mode, plus)
        self.RDB3 = ResidualDenseBlock5C(nc, kernel_size, gc, stride, bias, pad_type, norm_type, act_type, mode, plus)

    def forward(self, x):
        out = self.RDB1(x)
        out = self.RDB2(out)
        out = self.RDB3(out)
        # Empirically, we use 0.2 to scale the residual for better performance
        return out.mul(0.2) + x


def conv1x1(in_planes: int, out_planes: int, stride=1) -> nn.Conv2d:
    return nn.Conv2d(in_planes, out_planes, kernel_size=1, stride=stride, bias=False)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['vsgan', 'vsgan.archs', 'vsgan.networks']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.19.5,<2.0.0']

extras_require = \
{'docs': ['Sphinx>=4.3.2,<5.0.0',
          'furo>=2021.11.23,<2022.0.0',
          'dunamai>=1.7.0,<2.0.0',
          'm2r2>=0.3.2,<0.4.0',
          'sphinxcontrib-youtube>=1.0.1,<2.0.0',
          'sphinxcontrib-images>=0.9.4,<0.10.0'],
 'pytorch': ['torch>=1.10.0,<2.0.0'],
 'vs': ['VapourSynth>=57,<58']}

setup_kwargs = {
    'name': 'vsgan',
    'version': '1.6.0',
    'description': 'VapourSynth Single Image Super-Resolution Generative Adversarial Network (GAN)',
    'long_description': '![banner](https://rawcdn.githack.com/rlaphoenix/VSGAN/d7ad537bffb52bdbd1ad07c825cf016964ac57a2/banner.png)\n\n[![Build Tests](https://img.shields.io/github/workflow/status/rlaphoenix/VSGAN/Version%20test?label=Python%203.6%2B%20builds)](https://github.com/rlaphoenix/VSGAN/actions?query=workflow%3A%22Version+test%22)\n[![License](https://img.shields.io/github/license/rlaphoenix/VSGAN?style=flat)](https://github.com/rlaphoenix/VSGAN/blob/master/LICENSE)\n[![DeepSource](https://deepsource.io/gh/rlaphoenix/VSGAN.svg/?label=active+issues)](https://deepsource.io/gh/rlaphoenix/VSGAN/?ref=repository-badge)\n<a href="https://colab.research.google.com/github/rlaphoenix/VSGAN/blob/master/VSGAN.ipynb">\n    <img align="right" src="https://colab.research.google.com/assets/colab-badge.svg" alt="Open in Colab"/>\n</a>\n\nVSGAN is a Single Image Super-Resolution Generative Adversarial Network (GAN) which uses the VapourSynth processing framework to handle input and output image data.\n\n**Documentation**: https://vsgan.phoeniix.dev  \n**License**: [MIT License](LICENSE)\n',
    'author': 'PHOENiX',
    'author_email': 'rlaphoenix@pm.me',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/rlaphoenix/vsgan',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6.2,<4.0',
}


setup(**setup_kwargs)
